@@ -10,7 +10,7 @@
 #' names are found in tt. Supply them heirarchically (e.g. kingdom --> species)
 #' @param tax2map2 The taxonomic nomenclature you would like to map onto. pr2
 #' v4.12.0, Silva SSU v138 nr, GreenGenes v13.8 clustered at 97% similarity, and
-#' the RDP train set 16 are included in the ensembleTax package. You can map to
+#' the RDP train set v16 are included in the ensembleTax package. You can map to
 #' these by specifying "pr2", "Silva", "gg", or "rdp". Otherwise should be a
 #' dataframe of type character or list (no factors) with each column
 #' corresponding to a taxonomic rank.
@@ -22,9 +22,12 @@
 #' account for common variations in taxonomic name syntax and/or formatting
 #' commonly encountered in reference databases (e.g. Pseudo-nitzschia will map
 #' to Pseudonitzschia). If FALSE, formatting issues may preclude mapping of
-#' synonymous taxonomic names (e.g. Pseudonitzschia will NOT map to
-#' Pseudo-nitzschia). An exhaustive list of formatting details is included in
-#' Details.
+#' synonymous taxonomic names (e.g. Pseudo-nitzschia will NOT map to
+#' Pseudonitzschia). An exhaustive list of formatting details is included in
+#' Details. Note that formatting variants are only generated for the names in
+#' tt. This can cause some issues for mapping in the other direction (e.g.
+#' Pseudonitzschia in tt will NOT map to Pseudo-nitzschia in tax2map2 whether or
+#' not ignore.format is TRUE).
 #' @param synonym.file If "default", taxmapper uses taxonomic synonyms included
 #' with the ensembleTax package. If a custom taxonomic synonym file is
 #' preferred, a string corresponding to the name of the csv file should be
@@ -54,21 +57,40 @@
 #'
 #' When ignore.format = TRUE, names for which taxmapper cannot find exact
 #' matches in tax2map2 are altered in case an exact match was not found due to
-#' formatting issues. To do this taxmapper first checks for hyphens "-",
-#' underscores "_", and single spaces " ". If these are found, variants of the
-#' name with the hyphen/underscore/spaces replaced by each of the other two, as
-#' well as all subnames spearated by these characters, and all subnames pasted
-#' together with none of these special characters, are searched against tax2map2
-#' for exact matches. It also creates all-lower and all-upper case versions of
-#' these elements and again searches for exact name matches for these names.
-#' To prevent matching of arbitrary names often used in reference databases like
-#' "Clade X", after creating all of the above alternative names, those names
-#' that begin with any variant of the words "clade" or "group" and those names
-#' that are 2 characters or less are removed prior to re-searching tax2map2. All
+#' formatting issues. To do this taxmapper first removes square brackets ("[]").
+#' It then checks for hyphens "-", underscores "_", and single spaces " ". If
+#' these are found, variants of the name with the hyphen/underscore/spaces
+#' replaced by each of the other two, as well as all subnames spearated by these
+#' characters, and all subnames pasted together with none of these special
+#' characters, are searched against tax2map2 for exact matches. It also creates
+#' all-lower and all-upper case versions of these elements and again searches
+#' for exact name matches in tax2map2. Words generated by this process that are
+#' 2 characters or less are not searched for matches in tax2map2. All
 #' alternative names created when ignore.format = TRUE are also searched for
-#' synonyms in synonym.file. Be advised that setting ignore.format = TRUE does
-#' not guarantee a more finely resolved mapped taxonomy table, and can actually
-#' result in a less-resolved mapped taxonomy table in some circumstances.
+#' synonyms in synonym.file if specified.
+#'
+#' To prevent matching of arbitrary names often used in reference databases (eg,
+#' "Clade_X"), and after creating all of the above alternative names if
+#' ignore.format = TRUE, those that BEGIN with any of the words below are
+#' are not use in exact name matching. Instead, the lowest assigned
+#' non-ambiguous name is determined (any name that begins with a word NOT
+#' included in the list below) and is appended to the ambiguous name separated
+#' by a hyphen. The words taxmapper flags as ambiguous are: "Clade", "CLADE",
+#' "clade", "Group", "GROUP", "group", "Class", "CLASS", "class",
+#' "Subclass", "SubClass", "SUBCLASS", "subclass", "Subclade", "SubClade",
+#' "SUBCLADE", "subclade", "Subgroup", "SubGroup", "SUBGROUP", "subgroup",
+#' "Sub group", "Sub Group", "SUB GROUP", "sub group", "Sub clade", "Sub Clade",
+#' "SUB CLADE", "sub clade", "Sub class", "Sub Class", "SUB CLASS", "sub class",
+#' "Sub_group", "Sub_Group", "SUB_GROUP", "sub_group", "Sub_clade", "Sub_Clade",
+#' "SUB_CLADE", "sub_clade", "Sub_class", "Sub_Class", "SUB_CLASS", "sub_class",
+#' "Sub-group", "Sub-Group", "SUB-GROUP", "sub-group", "Sub-clade", "Sub-Clade",
+#' "SUB-CLADE", "sub-clade", "Sub-class", "Sub-Class", "SUB-CLASS", "sub-class",
+#' "incertae sedis", "INCERTAE SEDIS", "Incertae sedis", "Incertae Sedis",
+#' "incertae-sedis", "INCERTAE-SEDIS", "Incertae-sedis", "Incertae-Sedis",
+#' "incertae_sedis", "INCERTAE_-SEDIS", "Incertae_sedis", "Incertae_Sedis",
+#' "incertaesedis", "INCERTAESEDIS", "Incertaesedis", "IncertaeSedis",
+#' "unclassified", "UNCLASSIFIED", "Unclassified", "Novel", "novel", "NOVEL",
+#' "sp", "sp.", "spp", "spp.", "lineage", "Lineage", "LINEAGE"
 #'
 #' For high-throughput implementation of taxmapper, it's recommended to set
 #' streamline = TRUE.
@@ -137,21 +159,60 @@ taxmapper <- function(tt,
   }
   tax2map2.ranks <- colnames(tax2map2)
 
+  # function to test for ambiguous names (clade, group, subgroup, etc)
+  testAmbigNames <- function(taxonomy) {
+    # designate ambiguous names:
+    ambigu <- c("Clade", "CLADE", "clade", # variants of clade
+                "Group", "GROUP", "group", # variants of group
+                "Class", "CLASS", "class", # variants of class
+                "Subgroup", "SubGroup", "SUBGROUP", "subgroup", # subgroup
+                "Subclade", "SubClade", "SUBCLADE", "subclade", # subclade
+                "Subclass", "SubClass", "SUBCLASS", "subclass", # subclass
+                "Sub group", "Sub Group", "SUB GROUP", "sub group",
+                "Sub clade", "Sub Clade", "SUB CLADE", "sub clade",
+                "Sub class", "Sub Class", "SUB CLASS", "sub class",
+                "Sub_group", "Sub_Group", "SUB_GROUP", "sub_group",
+                "Sub_clade", "Sub_Clade", "SUB_CLADE", "sub_clade",
+                "Sub_class", "Sub_Class", "SUB_CLASS", "sub_class",
+                "Sub-group", "Sub-Group", "SUB-GROUP", "sub-group",
+                "Sub-clade", "Sub-Clade", "SUB-CLADE", "sub-clade",
+                "Sub-class", "Sub-Class", "SUB-CLASS", "sub-class",
+                "incertae sedis", "INCERTAE SEDIS", "Incertae sedis", "Incertae Sedis", # Incertae sedis
+                "incertae-sedis", "INCERTAE-SEDIS", "Incertae-sedis", "Incertae-Sedis", # more...
+                "incertae_sedis", "INCERTAE_-SEDIS", "Incertae_sedis", "Incertae_Sedis", # more...
+                "incertaesedis", "INCERTAESEDIS", "Incertaesedis", "IncertaeSedis",
+                "unclassified", "UNCLASSIFIED", "Unclassified",
+                "Novel", "novel", "NOVEL",
+                "sp", "sp.", "spp.", "spp",
+                "lineage", "Lineage", "LINEAGE")
+    # get start location of all ambiguous names within taxonomy (returns NA if not found)
+    loki <- stringr::str_locate(taxonomy, ambigu)[ , "start"]
+    test.val <- base::any(loki == 1) # TRUE if taxonomy starts with an ambiguous name. This means name should not be mapped
+    if (base::is.na(test.val)) { test.val <- FALSE } # account for NA's
+    return(test.val)
+  }
+
   # function to remove hyphens, underscores, upper case of name
   preprocessTax <- function(taxonomy) {
+    if (stringr::str_detect(taxonomy, c("\\[|\\]"))) {
+      taxonomy <- stringr::str_replace_all(taxonomy, c("\\[|\\]"), "")
+    }
     alt.full <- c(stringr::str_replace_all(taxonomy, "-"," "),
                   stringr::str_replace_all(taxonomy, "-","_"),
                   stringr::str_replace_all(taxonomy, " ","-"),
                   stringr::str_replace_all(taxonomy, " ","_"),
                   stringr::str_replace_all(taxonomy, "_","-"),
-                  stringr::str_replace_all(taxonomy, "_"," "))
+                  stringr::str_replace_all(taxonomy, "_"," "),
+                  stringr::str_replace_all(taxonomy, c("_|-"), " "),
+                  stringr::str_replace_all(taxonomy, c("_| "), "-"),
+                  stringr::str_replace_all(taxonomy, c(" |-"), "_"))
     # split terms by hyphens
     no.hyphen <- base::strsplit(taxonomy, "-")
     # split terms by underscores
     no.underscore <- base::strsplit(taxonomy, "_")
     # split by space:
     no.spc <- base::strsplit(taxonomy, " ")
-    # split terms by first instance of underscores and combine previous splits
+    # combine previous splits
     taxs <- c(taxonomy,
               alt.full,
               no.hyphen[[1]], no.underscore[[1]], no.spc[[1]],
@@ -165,25 +226,21 @@ taxmapper <- function(tt,
     no.lower <- base::toupper(taxs)
     # create alternative suffixes for certain taxonomies
     final.taxs <- unique(c(taxs, no.upper, no.lower))
-    # remove names that start with clade/group (and format variants), and/or
-    # short names as these are likely to be numbers
-    cl <- stringr::str_locate(final.taxs, "clade")
-    cl2 <- stringr::str_locate(final.taxs, "Clade")
-    cl3 <- stringr::str_locate(final.taxs, "CLADE")
-    gr <- stringr::str_locate(final.taxs, "group")
-    gr2 <- stringr::str_locate(final.taxs, "Group")
-    gr3 <- stringr::str_locate(final.taxs, "GROUP")
+    ambig.ones <- base::sapply(final.taxs, FUN = testAmbigNames)
     ll <- stringr::str_length(final.taxs)
-    rm.rows <- c(which(cl[, "start"] == 1),
-                 which(cl2[, "start"] == 1),
-                 which(cl3[, "start"] == 1),
-                 which(gr[, "start"] == 1),
-                 which(gr2[, "start"] == 1),
-                 which(gr3[, "start"] == 1),
+    rm.rows <- c(which(ambig.ones),
                  which(ll <= 2))
-    final.taxs <- final.taxs[-rm.rows]
-    # this ensures that the original name is always the first one mapped:
-    final.taxs <- c(taxonomy, final.taxs)
+    if (length(rm.rows) == 0) {
+      # do nothing to final.taxs
+    } else if (length(rm.rows > 0)) {
+      final.taxs <- final.taxs[-rm.rows]
+    }
+    # this ensures that the original name is always the first one mapped,
+    # and that longer names take priority over shorter names (thought being that
+    # longer names are more informative)
+    ii <- base::sort(stringr::str_length(final.taxs), decreasing = TRUE, index.return = TRUE)
+    ft.sorted <- final.taxs[ii$ix]
+    final.taxs <- c(taxonomy, ft.sorted)
     return(final.taxs)
   }
 
@@ -237,7 +294,7 @@ taxmapper <- function(tt,
     synonyms <- utils::read.csv(synonym.file, stringsAsFactors = FALSE)
     synonyms <- synonyms[, colnames(synonyms)[startsWith(colnames(synonyms), "Name")]]
   } else if (synonym.file == "default") {
-    synonyms <- ensembleTax::synonyms_20200816
+    synonyms <- ensembleTax::synonyms_v2
     synonyms <- synonyms[, colnames(synonyms)[startsWith(colnames(synonyms), "Name")]]
   }
 
@@ -265,15 +322,23 @@ taxmapper <- function(tt,
       for (col in 1:ncol(taxin.u)) {
         # keep track of the original taxonomy name
         orig.tax <- taxin.u[row, taxin.cols[col]]
+        # new addition: test for ambiguous names and make unambiguous if found
+        if (testAmbigNames(orig.tax) & !(is.na(orig.tax))) {
+          # if ambiguous name is found, grab the lowest annotated name that is not ambiguous
+          realnm.finder <- base::sapply(taxin.u[row , names(taxin.u)[base::seq(from = 1, to = (ncol(taxin.u) - col), by = 1)]], FUN = testAmbigNames)
+          realnm <- taxin.u[row , max(which(!(realnm.finder)))]
+          # then, append this name to the ambiguous name. This should make it unambiguous
+          orig.tax <- base::paste(realnm, orig.tax, sep = "-")
+        }
+
         # process the name to get alternatives by igorning its format
-        if (ignore.format) {
+        if (ignore.format & !(is.na(orig.tax))) {
           pos.taxs <- preprocessTax(orig.tax)
           for(tax in pos.taxs) {
             pos.taxs <- c(pos.taxs, getSynonyms(tax, synonyms))
           }
           pos.taxs <- base::unique(c(orig.tax, getSynonyms(orig.tax, synonyms), pos.taxs))
-        }
-        else {
+        } else {
           pos.taxs <- c(orig.tax, getSynonyms(orig.tax, synonyms))
         }
         # flag to keep track of when the row is already matched
@@ -281,21 +346,20 @@ taxmapper <- function(tt,
         # counter to keep track what column number we are on
         counter <- 1
         # iterate through all alternatives of the taxonomy name with original one first
-        for (taxonomy in pos.taxs) {
+        for (tax2map in pos.taxs) {
           last <- FALSE
           if (counter == length(pos.taxs)) {
             last <- TRUE
           }
-          if (!is.na(taxonomy)) {
+          if (!is.na(tax2map)) {
             # find matching
-            match <- findMapping(taxonomy, tax2map2.u)
+            match <- findMapping(taxonomy = tax2map, tax2map2 = tax2map2.u)
             if (is.data.frame(match)) {
               combined <- cbind(taxin.u[row, ], match)
               mapped <- rbind(mapped, combined)
               matched <- TRUE
               break
-            }
-            else { # if no matching is found, add to not.mapped
+            } else { # if no matching is found, add to not.mapped
               if (last) {
                 not.mapped <- c(not.mapped, orig.tax)
               }
